/*
 * Remote control driver for the TV-card
 * key codes are obtained from GPIO port
 *
 * For saa7134, based on version for bttv
 *
 * (L) by Artur Lipowski <alipowski@kki.net.pl>
 *     patch for the AverMedia by Santiago Garcia Mantinan <manty@i.am>
 *                            and Christoph Bartelmus <lirc@bartelmus.de>
 *     patch for the BestBuy by Miguel Angel Alvarez <maacruz@navegalia.com>
 *     patch for the Winfast TV2000 by Juan Toledo <toledo@users.sourceforge.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: lirc_gpio.c,v 1.16 2002/08/30 11:57:28 lirc Exp $
 *
 */

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 2, 4)
#error "*******************************************************"
#error "Sorry, this driver needs kernel version 2.2.4 or higher"
#error "*******************************************************"
#endif

#include <linux/module.h>
#include <linux/kmod.h>
#include <linux/sched.h>
#include <linux/wrapper.h>
#include <linux/errno.h>

#include "include/saa7134.h"
#include "include/saa7134-gpio.h"
#include "../lirc_dev/lirc_dev.h"

#define dprintk  if (debug) printk

static int debug = 0;
static int card = 0;
static int minor = -1;
static unsigned long gpio_mask = 0;
static unsigned long gpio_enable = 0;
static unsigned long gpio_lock_mask = 0;
static unsigned long gpio_xor_mask = 0;
static unsigned int soft_gap = 0;
static unsigned char sample_rate = 10;

MODULE_PARM(debug,"i");
MODULE_PARM(card,"i");
MODULE_PARM(minor,"i");
MODULE_PARM(gpio_mask,"l");
MODULE_PARM(gpio_lock_mask,"l");
MODULE_PARM(gpio_xor_mask,"l");
MODULE_PARM(soft_gap,"i");
MODULE_PARM(sample_rate,"b");

static unsigned char code_length = 0;
static unsigned char code_bytes = 1;

struct rcv_info {
	int board_id;
	unsigned long pci_subsystem_id;
	unsigned long gpio_mask;
	unsigned long gpio_enable;
	unsigned long gpio_lock_mask;
	unsigned long gpio_xor_mask;
	unsigned int soft_gap;
	unsigned char sample_rate;
	unsigned char code_length;
};

static struct rcv_info rcv_infos[] = {
	{ SAA7134_BOARD_FLYVIDEO3000,	0, 0xEE00000, 0, 0x0040000, 0x0040000, 0, 10, 0 }
/* 	{BTTV_FLYVIDEO_98,            0, 0x000001f8,          0, 0x0000100,          0,   0,  0,  0},
        {BTTV_WINFAST2000,            0, 0x000000f8,          0, 0x0000100,          0,   0,  0,  0}
*/
};

static int board_id = 0;

#define MAX_BYTES 8

#define SUCCESS 0
#define LOGHEAD "lirc_gpio_s (%d): "

/* how many bits GPIO value can be shifted right before processing
 * it is computed from the value of gpio_mask_parameter
 */

static unsigned char gpio_pre_shift = 0;

static inline int reverse(int data, int bits)
{
	int i;
	int c;

	for (c=0,i=0; i<bits; i++) {
		c |= (((data & (1<<i)) ? 1:0)) << (bits-1-i);
	}

	return c;
}

static int build_key(unsigned long gpio_val, unsigned char codes[MAX_BYTES])
{
	unsigned char shift = 0;
	unsigned long mask = gpio_mask;

	dprintk(LOGHEAD "gpio_val is %lx\n",card,(unsigned long) gpio_val);

	gpio_val ^= gpio_xor_mask;

	if (gpio_lock_mask && (gpio_val & gpio_lock_mask)) {
		return -EBUSY;
	}

	/* extract bits from "raw" GPIO value using gpio_mask */
	codes[0] = 0;
	gpio_val >>= gpio_pre_shift;
	while (mask) {
		if (mask & 1u) {
			codes[0] |= ((gpio_val & 1u) << shift);
			shift++;
		}
		mask >>= 1;
		gpio_val >>= 1;
	}

	switch (board_id)
	{
	case SAA7134_BOARD_FLYVIDEO3000:
		codes[0]=reverse(codes[0],6);
		break;
	default:
		break;
	}

	dprintk(LOGHEAD "code is %lx\n",card,(unsigned long) codes[0]);

	return SUCCESS;
}

static int get_key(void* data, unsigned char *key, int key_no)
{
	static unsigned long next_time = 0;
	static unsigned char codes[MAX_BYTES];
	unsigned long code = 0;
	unsigned char cur_codes[MAX_BYTES];

	if (key_no > 0)	{
		if (code_bytes < 2 || key_no >= code_bytes) {
			dprintk(LOGHEAD "something wrong in get_key\n",card);
			return -EBADRQC;
		}
		*key = codes[key_no];
		return SUCCESS;
	}

	if (saa7134_read_gpio(card, &code)) {
		dprintk(LOGHEAD "cannot read GPIO\n",card);
		return -EIO;
	}

	if (build_key(code, cur_codes)) {
		return -EFAULT;
	}

	if (soft_gap) {
		if (!memcmp(codes, cur_codes, code_bytes) &&
		    jiffies < next_time) {
			return -EAGAIN;
		}
		next_time = jiffies + soft_gap;
	}

	memcpy(codes, cur_codes, code_bytes);

	*key = codes[0];

	return SUCCESS;
}

static void set_use_inc(void* data)
{
	MOD_INC_USE_COUNT;
}

static void set_use_dec(void* data)
{
	MOD_DEC_USE_COUNT;
}

static wait_queue_head_t* get_queue(void* data)
{
	return saa7134_get_gpio_queue(card);
}

static struct lirc_plugin plugin = {
	"lirc_gpio_s",
	0,
	0,
	0,
	NULL,
	get_key,
	get_queue,
	set_use_inc,
	set_use_dec
};

/*
 *
 */
int gpio_remote_init(void)
{  	
	int ret;
	unsigned int mask;

	/* "normalize" gpio_mask
	 * this means shift it right until first bit is set
	 */
	while (!(gpio_mask & 1u)) {
		gpio_pre_shift++;
		gpio_mask >>= 1;
	}

	if (code_length) {
		plugin.code_length = code_length;
	} else {
		/* calculate scan code length in bits if needed */
		plugin.code_length = 1;
		mask = gpio_mask >> 1;
		while (mask) {
			if (mask & 1u) {
				plugin.code_length++;
			}
			mask >>= 1;
		}
	}

	code_bytes = (plugin.code_length/8) + (plugin.code_length%8 ? 1 : 0);
	if (MAX_BYTES < code_bytes) {
		printk (LOGHEAD "scan code too long (%d bytes)\n",
			minor, code_bytes);
		return -EBADRQC;
	}

	/* translate ms to jiffies */
	soft_gap = (soft_gap*HZ) / 1000;

	plugin.minor = minor;
	plugin.sample_rate = sample_rate;

	ret = lirc_register_plugin(&plugin);
	
	if (0 > ret) {
		printk (LOGHEAD "device registration failed with %d\n",
			minor, ret);
		return ret;
	}

	minor = ret;
	printk(LOGHEAD "driver registered\n", minor);

	return SUCCESS;
}

EXPORT_NO_SYMBOLS; 

/* Dont try to use it as a static version !  */

#ifdef MODULE
MODULE_DESCRIPTION("Driver module for remote control (data from saa7134 GPIO port)");
MODULE_AUTHOR("Artur Lipowski");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

/*
 *
 */
int init_module(void)
{
	unsigned long pci_subsystem_id;
	int rcv_index;

	if (MAX_IRCTL_DEVICES < minor) {
		printk("lirc_gpio_s: parameter minor (%d) must be less than %d!\n",
		       minor, MAX_IRCTL_DEVICES-1);
		return -EBADRQC;
	}

	request_module("saa7134");

	saa7134_get_cardinfo(card, &board_id, &pci_subsystem_id);

	printk(LOGHEAD "card saa7134[%u] board %u, pci subsystem 0x%x\n",minor,card,board_id,pci_subsystem_id);

	if(gpio_mask == 0) { /* if gpio_mask is 0, automatically set values */
		for (rcv_index = 0; rcv_index < sizeof(rcv_infos)/sizeof(struct rcv_info); rcv_index++) {
			if (rcv_infos[rcv_index].board_id == board_id &&
			    (rcv_infos[rcv_index].pci_subsystem_id == 0 ||
			     rcv_infos[rcv_index].pci_subsystem_id == pci_subsystem_id)) {
				gpio_mask = rcv_infos[rcv_index].gpio_mask;
				gpio_enable = rcv_infos[rcv_index].gpio_enable;
				gpio_lock_mask = rcv_infos[rcv_index].gpio_lock_mask;
				gpio_xor_mask = rcv_infos[rcv_index].gpio_xor_mask;
				soft_gap = rcv_infos[rcv_index].soft_gap;
				sample_rate = rcv_infos[rcv_index].sample_rate;
				code_length = rcv_infos[rcv_index].code_length;
				}
		}
	}
	else {
		printk(LOGHEAD "TV card type %u not supported! Manually provide values.\n",minor,board_id);
		return -EBADRQC;
	}

	/* Check parameters */
	if (sample_rate < 2 || sample_rate > 50) {
		printk(LOGHEAD "parameter sample_rate must be beetween 2 and 50!\n", minor);
		return -EBADRQC;
	}

	if (soft_gap && ((2000/sample_rate) > soft_gap || 1000 < soft_gap)) {
		printk(LOGHEAD "parameter soft_gap must be beetween %d and 1000!\n",
			minor, 2000/sample_rate);
		return -EBADRQC;
	}

	request_module("lirc_dev");

	return gpio_remote_init();
}

/*
 *
 */
void cleanup_module(void)
{
	int ret;

	ret = lirc_unregister_plugin(minor);

	if (0 > ret) {
		printk(LOGHEAD "error in lirc_unregister_minor: %d\n"
		       "Trying again...\n",
		       minor, ret);

		current->state = TASK_INTERRUPTIBLE;
		schedule_timeout(HZ);

		ret = lirc_unregister_plugin(minor);

		if (0 > ret) {
			printk(LOGHEAD "error in lirc_unregister_minor: %d!!!\n",
			       minor, ret);
			return;
		}
	}

	dprintk(LOGHEAD "module successfully unloaded\n", minor);
}
#endif

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
